%{
#include "assembleur_ksi.h"
#include "consts.h"
#include <stdio.h>

static char *mnemoniques[] = {
  NULL, "AFFECTE", "AFFECTE+", "SI", "SAUT", "CPILE", "DPILE", "APPEL",
  "RETOUR", "TEST", "TRAPPE" , "AFFECTESP" };

static char *trappecode[] = {
  NULL, "CLONE", "RECOUVRE", "ATTENDS", "FIN", "CAPTURE", "EMETS", "ID",
  "IDP", "LIT", "ECRIT" };

static inlabels(const char *);
 static void add_unsolved(char *,char);

static ins_t tab_ins[MAX_INS];
static int current_instruction = 0;
static int current_label = 0;
static int current_unsolved_label = 0;
static int donnees_taille = 0;
static struct {
       char label[LABEL_MAX+1];
       int           instruction_num;
} labels[MAX_LABELS];
extern int numligne;
static int l;
static struct {
       char label[LABEL_MAX+1];
       char valeur;
       int instruction_num;
} unsolved_labels[MAX_LABELS];

%}

%union {
  ins_t           instruction;
  char   valeur;
  char   reference[2];
  char   label[LABEL_MAX+1];
}

%token <valeur> TOK_DONNEES
%token <valeur> TOK_AFFECTE
%token <valeur> TOK_SAUT
%token <valeur> TOK_PILE
%token <valeur> TOK_APPEL
%token <valeur> TOK_RETOUR
%token <valeur> TOK_TEST
%token <valeur> TOK_TRAPPE
%token <valeur> TOK_CTRAPPE
%token SEPARATOR
%token EOL
%token <reference> REFERENCE
%token <reference> CONSTANTE
%token <label> LABEL
%type <instruction> inst
%type <instruction> inst_affecte
%type <instruction> inst_affectesp
%type <instruction> inst_pile
%type <instruction> inst_retour
%type <instruction> inst_trappe
%type <instruction> inst_appel
%type <instruction> inst_test
%type <instruction> inst_saut
%type <reference> ref_or_cst
%%
program :
         SEPARATOR donnees EOL insts_list;
insts_list :
	   inst_or_label EOL
	 | insts_list inst_or_label EOL
	 ;

donnees : TOK_DONNEES SEPARATOR CONSTANTE
        {
	  donnees_taille = $3[1];
        }

inst_or_label : SEPARATOR inst
        {
	  tab_ins[current_instruction++] = $2;
	}
       | label ':'
	 ;

label : LABEL
           {
	     if (inlabels($1)==-1) {
	       strcpy(labels[current_label].label,$1);
	       labels[current_label++].instruction_num = current_instruction;
	     }
	     else {
	       fprintf(stderr,"Redefinition de label en ligne %d (%s)\n",
		       numligne,$1);
	       exit(1);
	     }
	   }
         ;

inst : inst_affecte
       | inst_affectesp
       | inst_saut
       | inst_pile
       | inst_appel
       | inst_retour
       | inst_test
       | inst_trappe
	 ;

inst_affecte : TOK_AFFECTE SEPARATOR REFERENCE ',' ref_or_cst
                {
		  SET_INS($$,$1);
		  SET_TYPE1($$,$3[0]);
		  SET_PAR1($$,$3[1]);
		  SET_TYPE2($$,$5[0]);
		  SET_PAR2($$,$5[1]);
		}
         ;
inst_affectesp : TOK_AFFECTE SEPARATOR REFERENCE ',' LABEL
                {
		  SET_INS($$,$1);
		  SET_TYPE1($$,$3[0]);
		  SET_PAR1($$,$3[1]);
		  SET_TYPE2($$,0);
		  if ( (l=inlabels($5)) != -1 ) {
		    SET_PAR2($$,labels[l].instruction_num);
		  } else {
		    add_unsolved($5,$1);
		  }
		}
         ;
inst_retour : TOK_RETOUR
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,0);
	    SET_PAR1($$,0);
	    SET_TYPE2($$,0);
	    SET_PAR2($$,0);
	  }
inst_trappe : TOK_TRAPPE SEPARATOR TOK_CTRAPPE
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,0);
	    SET_PAR1($$,$3);
	    SET_TYPE2($$,0);
	    SET_PAR2($$,0);
	  }
inst_saut : TOK_SAUT SEPARATOR LABEL
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,0);
	    if ( (l=inlabels($3)) != -1 ) {
	      SET_PAR1($$,labels[l].instruction_num);
	    } else {
	      add_unsolved($3,$1);
	    }
	    SET_TYPE2($$,0);
	    SET_PAR2($$,0);
	  }
inst_pile : TOK_PILE SEPARATOR CONSTANTE
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,$3[0]);
	    SET_PAR1($$,$3[1]);
	    SET_TYPE2($$,0);
	    SET_PAR2($$,0);
	  }
inst_appel : TOK_APPEL SEPARATOR LABEL
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,0);
	    if ( (l=inlabels($3)) != -1 ) {
	      SET_PAR1($$,labels[l].instruction_num);
	    } else {
	      add_unsolved($3,$1);
	    }
	    SET_TYPE2($$,0);
	    SET_PAR2($$,0);
	  }
inst_test : TOK_TEST SEPARATOR ref_or_cst ',' ref_or_cst
          {
	    SET_INS($$,$1);
	    SET_TYPE1($$,$3[0]);
	    SET_PAR1($$,$3[1]);
	    SET_TYPE2($$,$5[0]);
	    SET_PAR2($$,$5[1]);
	  }
ref_or_cst :
	   REFERENCE
         | CONSTANTE
         ;
%%
static inlabels(const char *label) {
  int i;

  for (i=0; i<current_label; i++)
    if (!strcmp(label,labels[i].label)) return i;
  return -1;
}

static void add_unsolved(char *label,char ins) {
  if (current_unsolved_label==MAX_LABELS) {
    fprintf(stderr,"Impossible de resoudre les etiquettes en avant\n");
    exit(1);
  }
  strcpy(unsolved_labels[current_unsolved_label].label,label);
  unsolved_labels[current_unsolved_label].valeur = ins;
  unsolved_labels[current_unsolved_label++].instruction_num = current_instruction;
}

void printref1(ins_t ins) {
  int t;

  t = GET_TYPE1(ins);
  if (t==0) {
    fprintf(stderr,"#%d",GET_PAR1(ins));
    return;
  }
  if (t&0x01)
    fprintf(stderr,"*");
  if (t&0x02)
    fprintf(stderr,"P");
  else
    fprintf(stderr,"M");
  fprintf(stderr,"%d",GET_PAR1(ins));
}

void printref2(ins_t ins) {
  int t;

  t = GET_TYPE2(ins);
  if (t==0) {
    fprintf(stderr,"#%d",GET_PAR2(ins));
    return;
  }
  if (t&0x01)
    fprintf(stderr,"*");
  if (t&0x02)
    fprintf(stderr,"P");
  else
    fprintf(stderr,"M");
  fprintf(stderr,"%d",GET_PAR2(ins));
}

void printaddr2(ins_t ins) {
  fprintf(stderr,"i%03d",GET_PAR2(ins));
}

void printaddr1(ins_t ins) {
  fprintf(stderr,"i%03d",GET_PAR1(ins));
}

void printaffecte(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printref1(ins);
  fprintf(stderr,",");
  printref2(ins);
  fprintf(stderr,"\n");
}

void printaffectesp(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printref1(ins);
  fprintf(stderr,",");
  printaddr2(ins);
  fprintf(stderr,"\n");
}

void printsi(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printaddr1(ins);
  fprintf(stderr,"\n");
}

void printsaut(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printaddr1(ins);
  fprintf(stderr,"\n");
}

void printtest(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printref1(ins);
  fprintf(stderr,",");
  printref2(ins);
  fprintf(stderr,"\n");
}

void printappel(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printaddr1(ins);
  fprintf(stderr,"\n");
}

void printretour(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  fprintf(stderr,"\n");
}

void printtrappe(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  fprintf(stderr,"%s",trappecode[GET_PAR1(ins)]);
  fprintf(stderr,"\n");
}

void printcpile(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printref1(ins);
  fprintf(stderr,"\n");
}

void printdonnees(ins_t ins) {
  fprintf(stderr,"%10s ","DONNEES");
  printref1(ins);
  fprintf(stderr,"\n");
}

void printdpile(ins_t ins) {
  fprintf(stderr,"%10s ",mnemoniques[ins>>24]);
  printref1(ins);
  fprintf(stderr,"\n");
}

int main(int argc,char *argv[]) {
  int i, l;
    ins_t *ins;
    int old_current;

    current_instruction = 4;
    yyparse();
    old_current = current_instruction;
    for (i=0; i<current_unsolved_label; i++) {
      if ((l=inlabels(unsolved_labels[i].label))==-1) {
	fprintf(stderr,"Label inconnu %s\n",unsolved_labels[i].label);
	exit(1);
      }
      ins = tab_ins+unsolved_labels[i].instruction_num;
      switch(unsolved_labels[i].valeur) {
      case SI:
      case SAUT:
      case APPEL:
	SET_PAR1(*ins,labels[l].instruction_num);
	break;
      case AFFECTESP:
	SET_PAR2(*ins,labels[l].instruction_num);
	break;
      default:
	fprintf(stderr,"Problemes sur les labels\n");
	exit(1);
      }
    }

    current_instruction = 0;
    SET_INS(tab_ins[current_instruction],DONNEES);
    SET_TYPE1(tab_ins[current_instruction],0);
    SET_PAR1(tab_ins[current_instruction],donnees_taille);
    current_instruction++;

    // ins: CPILE #1
    SET_INS(tab_ins[current_instruction],CPILE);
    SET_TYPE1(tab_ins[current_instruction],0);
    SET_PAR1(tab_ins[current_instruction],1);
    current_instruction++;

    // ins: APPEL debut ou +3
    if ((l=inlabels("debut"))==-1) {
      fprintf(stderr,"Attention il manque l'etiquette de debut\n");
      fprintf(stderr,"  ce code demarre donc a la premiere instruction!\n");
      SET_INS(tab_ins[current_instruction],APPEL);
      SET_TYPE1(tab_ins[current_instruction],0);
      SET_PAR1(tab_ins[current_instruction],current_instruction+3);
    }
    else {
      SET_INS(tab_ins[current_instruction],APPEL);
      SET_TYPE1(tab_ins[current_instruction],0);
      SET_PAR1(tab_ins[current_instruction],labels[l].instruction_num);
    }
    current_instruction++;

    // ins: TRAPPE FIN
    SET_INS(tab_ins[current_instruction],TRAPPE);
    SET_TYPE1(tab_ins[current_instruction],0);
    SET_PAR1(tab_ins[current_instruction],FIN);

    write(1,tab_ins,old_current*sizeof(ins_t));
    for (i=0; i<old_current; i++) {
      fprintf(stderr,"i%03d: ",i);
      switch(tab_ins[i]>>24) {
      case AFFECTE:
      case AFFECTEP:
	printaffecte(tab_ins[i]);
	break;
      case AFFECTESP:
	printaffectesp(tab_ins[i]);
	break;
      case SI:
	printsi(tab_ins[i]);
	break;
      case SAUT:
	printsaut(tab_ins[i]);
	break;
      case RETOUR:
	printretour(tab_ins[i]);
	break;
      case APPEL:
	printappel(tab_ins[i]);
	break;
      case TRAPPE:
	printtrappe(tab_ins[i]);
	break;
      case CPILE:
	printcpile(tab_ins[i]);
	break;
      case DPILE:
	printdpile(tab_ins[i]);
	break;
      case TEST:
	printtest(tab_ins[i]);
	break;
      case DONNEES:
	printdonnees(tab_ins[i]);
	break;
      default:
	fprintf(stderr,"       UNK\n");
	break;
      }
    }
    exit(0);
}
